from typing import Any, Dict, List, Optional, Tuple

from motor.motor_asyncio import AsyncIOMotorClient

from chatushka.core.services.mongodb.wrapper import MongoDBWrapper


def get_mongodb_client() -> AsyncIOMotorClient:
    return MongoDBWrapper().client


# def get_database() -> AsyncIOMotorCollection:
#     client = get_mongodb_client()
#     return client[settings.mongodb.database]


async def mongodb_paginated_find(
    collection: AsyncIOMotorClient,
    query: Optional[Dict[str, Any]],
    page: int,
    per_page: int,
    sort: Optional[List[Tuple[str, int]]] = None,
    projection: Optional[Dict[str, int]] = None,
) -> List[Dict[str, Any]]:
    cursor = collection.find(query, sort=sort, projection=projection)
    cursor.skip(page * per_page).limit(per_page)
    return [doc async for doc in cursor]



from chatushka.core.utils import ServiceSettingsBase


class MongoDBSettings(ServiceSettingsBase):
    mongodb_dsn: str
    mongodb_min_connections_count: int = 2
    mongodb_max_connections_count: int = 8



from datetime import datetime, timezone
from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field


class ChatType(str, Enum):
    GROUP = "group"
    SUPERGROUP = "supergroup"
    PRIVATE = "private"


class ChatMemberStatuses(str, Enum):
    MEMBER = "member"
    CREATOR = "creator"
    ADMINISTRATOR = "administrator"
    LEFT = "left"
    KICKED = "kicked"
    RESTRICTED = "restricted"


class User(BaseModel):
    id: int
    is_bot: bool
    first_name: Optional[str]
    last_name: Optional[str]
    can_join_groups: Optional[bool]
    can_read_all_group_messages: Optional[bool]

    @property
    def readable_name(self) -> str:
        return f"{self.first_name}{' '+self.last_name if self.last_name else ''}"


class ChatMemberBase(BaseModel):
    status: ChatMemberStatuses
    user: User
    is_anonymous: Optional[bool] = None
    custom_title: Optional[str] = None


class ChatMemberOwner(ChatMemberBase):
    ...


class ChatMemberAdministrator(ChatMemberBase):
    can_be_edited: Optional[bool]
    can_manage_chat: Optional[bool]
    can_delete_messages: Optional[bool]
    can_manage_voice_chats: Optional[bool]
    can_restrict_members: Optional[bool]
    can_promote_members: Optional[bool]
    can_change_info: Optional[bool]
    can_invite_users: Optional[bool]
    can_post_messages: Optional[bool]
    can_edit_messages: Optional[bool]
    can_pin_messages: Optional[bool]


class Chat(BaseModel):
    id: int
    type: ChatType
    title: Optional[str]


class NewChatMember(
    ChatMemberAdministrator,
):
    ...


class OldChatMember(
    ChatMemberAdministrator,
):
    ...


class Message(BaseModel):
    message_id: int
    user: User = Field(..., alias="from")
    chat: Chat
    text: Optional[str]
    reply_to_message: Optional["Message"] = None
    new_chat_members: list[User] = Field(default_factory=list)


Message.update_forward_refs()


class MyChatMember(BaseModel):
    chat: Chat
    user: User = Field(..., alias="from")
    date: datetime
    old_chat_member: OldChatMember
    new_chat_member: NewChatMember

    class Config:
        json_decoders = dict(datetime=lambda v: datetime.fromtimestamp(v, tz=timezone.utc))


class Update(BaseModel):
    update_id: int
    message: Optional[Message] = None
    my_chat_member: Optional[MyChatMember] = None


class ChatPermissions(BaseModel):
    can_send_messages: bool
    can_send_media_messages: bool
    can_send_polls: bool
    can_send_other_messages: bool
    # can_add_web_page_previews: bool
    # can_change_info: bool
    # can_invite_users: bool
    # can_pin_messages: bool



from logging import getLogger

from motor.motor_asyncio import AsyncIOMotorClient

from chatushka.core.services.base import ServiceWrapperBase
from chatushka.core.services.mongodb.settings import MongoDBSettings

logger = getLogger(__name__)


class MongoDBWrapper(ServiceWrapperBase):
    def __init__(self) -> None:
        super().__init__()
        self.healthz_name = "mongodb"
        self.client: AsyncIOMotorClient
        self.settings: MongoDBSettings = MongoDBSettings()

    async def startup_event_handler(
        self,
    ) -> None:
        self.client = AsyncIOMotorClient(  # noqa
            self.settings.mongodb_dsn,
            minPoolSize=self.settings.mongodb_min_connections_count,
            maxPoolSize=self.settings.mongodb_max_connections_count,
            uuidRepresentation="standard",
        )

    async def shutdown_event_handler(self) -> None:
        if self.client:
            self.client.close()

    async def health_check(self) -> None:
        await self.client.conn.command({"dbStats": 1})



from datetime import datetime, timedelta, timezone
from enum import Enum
from random import choice

from chatushka.bot.settings import get_settings
from chatushka.core.matchers import CommandsMatcher
from chatushka.core.transports.models import ChatPermissions, Message, User
from chatushka.core.transports.telegram_bot_api import TelegramBotApi

RESTRICT_PERMISSION = ChatPermissions(
    can_send_messages=False,
    can_send_media_messages=False,
    can_send_polls=False,
    can_send_other_messages=False,
)
settings = get_settings()
mute_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)


class MuteMessages(Enum):
    ACCIDENT = (
        '🧐 <a href="tg://user?id={user}">{name}</a> споткнулся, упал и попал в больницу на {time} минут.',
        '🧐 <a href="tg://user?id={user}">{name}</a> заигрался револьвером и угодил в травмпункт на {time} минут.',
        '🧐 <a href="tg://user?id={user}">{name}</a> переводил бабушку через дорогу и теперь отдыхает {time} минут.',
        '🧐 <a href="tg://user?id={user}">{name}</a> решил подумать о жизни {time} минут.',
    )
    LOOSER = (
        '🧐 <a href="tg://user?id={looser_id}">{looser_name}</a> хотел убить '
        '<a href="tg://user?id={victim_id}">{victim_name}</a>, но что-то пошло не так и он '
        "вынужден провести в тюрьме {time} минут",
        '🧐 У <a href="tg://user?id={looser_id}">{looser_name}</a> лапки коротковаты '
        'чтоб убить <a href="tg://user?id={victim_id}">{victim_name}</a>',
    )


async def send_mute_request(
    api: TelegramBotApi,
    message: Message,
    initiator: User,
    restrict_user: User,
    restrict_time: timedelta,
) -> None:
    text_tmpl = choice(MuteMessages.ACCIDENT.value)
    is_success = await api.restrict_chat_member(
        chat_id=message.chat.id,
        user_id=restrict_user.id,
        permissions=RESTRICT_PERMISSION,
        until_date=datetime.now(tz=timezone.utc) + restrict_time,
    )
    if is_success:
        await api.send_message(
            chat_id=message.chat.id,
            text=text_tmpl.format(
                user=restrict_user.id,
                name=restrict_user.readable_name,
                time=int(restrict_time.total_seconds() // 60),
            ),
        )
        return
    text_tmpl = choice(MuteMessages.LOOSER.value)
    await api.send_message(
        chat_id=message.chat.id,
        text=text_tmpl.format(
            looser_id=initiator.id,
            looser_name=initiator.readable_name,
            victim_id=restrict_user.id,
            victim_name=restrict_user.readable_name,
        ),
        reply_to_message_id=message.message_id,
    )



from typing import Any, Callable, Coroutine, Hashable, Iterable, Optional, Protocol, Union

from chatushka.core.models import HANDLER_TYPING, MatchedToken
from chatushka.core.transports.models import Message
from chatushka.core.transports.telegram_bot_api import TelegramBotApi


class MatcherProtocol(Protocol):

    handlers: dict[Hashable, list[HANDLER_TYPING]]

    def __call__(
        self,
        *tokens: Hashable,
    ) -> Callable[[Callable[[], None]], None]:
        ...

    def add_handler(
        self,
        tokens: Union[Hashable, Iterable[Hashable]],
        handler: Union[Callable[[], None], Callable[[], Coroutine]],  # type: ignore
    ) -> None:
        ...

    async def check_handlers(
        self,
        message: Message,
    ) -> None:
        ...

    async def match(
        self,
        api: TelegramBotApi,
        message: Message,
    ) -> MatchedToken:
        ...

    async def call(
        self,
        api: TelegramBotApi,
        token: Hashable,
        message: Optional[Message] = None,
        kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        ...

    async def init(self) -> None:
        ...



from functools import lru_cache
from pathlib import Path
from typing import Union

from chatushka.core.utils import ServiceSettingsBase

BOBUK_JOKES_URL = "https://jokesrv.rubedo.cloud/"

BOT_ROOT_DIR = Path(__file__).parent.resolve()
BOT_DATA_DIR = BOT_ROOT_DIR / "data"


class _Settings(ServiceSettingsBase):
    command_prefixes: Union[str, tuple[str, ...]] = ("/", "!")
    command_postfixes: Union[str, tuple[str, ...]] = "!"


@lru_cache
def get_settings() -> _Settings:
    return _Settings()



from abc import ABC, abstractmethod
from asyncio import wait_for
from logging import getLogger

from pydantic import BaseSettings

from chatushka import ChatushkaBot
from chatushka.core.models import EventTypes
from chatushka.core.utils import SingletonABCMeta

logger = getLogger(__name__)


class ServiceWrapperBase(ABC, metaclass=SingletonABCMeta):
    def __init__(self, healthz_timeout: int = 8):
        logger.debug(f"{self.__class__.__name__} instance created")
        self.healthz_timeout: int = healthz_timeout
        self.healthz_name: str = "service"
        self.settings: BaseSettings

    def add_event_handlers(
        self,
        bot: ChatushkaBot,
    ) -> None:
        bot.add_handler(EventTypes.STARTUP, self.startup_event_handler)
        bot.add_handler(EventTypes.SHUTDOWN, self.shutdown_event_handler)

    @abstractmethod
    async def startup_event_handler(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def shutdown_event_handler(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def health_check(self) -> None:
        raise NotImplementedError

    async def health_check_call(self) -> bool:
        is_passed = False
        try:
            fut = self.health_check()
            if self.healthz_timeout:
                await wait_for(fut, timeout=self.healthz_timeout)
            else:
                await fut
            is_passed = True
        except TimeoutError:
            logger.critical(f"Exceeded the timeout of {self.healthz_timeout} seconds", exc_info=True)
        except Exception:  # noqa  # pylint: disable=broad-except
            logger.critical("Error occurred", exc_info=True)

        return is_passed



from abc import ABCMeta
from typing import Any

from pydantic import BaseSettings


class ServiceSettingsBase(BaseSettings):
    class Config:
        env_prefix = "BOT_"
        env_file = ".env"
        allow_mutation = False


class SingletonABCMeta(ABCMeta):
    _instances = {}  # type: ignore

    def __call__(cls, *args: Any, **kwargs: Any):  # type: ignore
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonABCMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from datetime import datetime
from logging import getLogger
from typing import Any, Dict, List, Optional, Tuple, Union

from httpx import AsyncClient, Response
from pydantic import ValidationError

from chatushka.core.transports import models
from chatushka.core.transports.models import (
    ChatMemberAdministrator,
    ChatMemberOwner,
    ChatMemberStatuses,
    ChatPermissions,
)

logger = getLogger()


class TelegramBotApi:
    def __init__(
        self,
        token: str,
    ) -> None:
        self.token = token

    @property
    def _base_api_url(self) -> str:
        return f"https://api.telegram.org/bot{self.token}"

    def _api_method_url(
        self,
        method: str,
    ) -> str:
        return f"{self._base_api_url}/{method}"

    @staticmethod
    def check_api_response(
        response: Response,
    ) -> Dict[str, Any]:
        data: dict[str, Any] = response.json()
        is_ok: bool = data.get("ok", False)
        if not is_ok:
            logger.warning(response.text)
            raise ValueError(f"Telegram response error: {response.text}\n{data}")

        result: Dict[str, Any] = data["result"]
        return result

    async def _call_api(
        self,
        method: str,
        timeout: int = 10,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        url = self._api_method_url(method)
        async with AsyncClient() as client:  # type: AsyncClient
            response = await client.post(url, timeout=timeout * 2, data=kwargs)
        return self.check_api_response(response)

    async def get_me(
        self,
    ) -> models.User:
        result = await self._call_api("getme")
        return models.User(**result)  # type: ignore

    async def get_updates(
        self,
        timeout: int,
        offset: Optional[int] = None,
    ) -> Tuple[List[models.Update], int]:
        params = {}
        if offset:
            params["offset"] = offset
        results = await self._call_api(
            "getupdates",
            timeout=timeout,
            **params,
        )
        updates_list = []
        latest_update_id: Optional[int] = offset
        for result in results:
            if not latest_update_id or (latest_update_id < result["update_id"]):  # type: ignore
                latest_update_id = result["update_id"]  # type: ignore
            try:
                update = models.Update(**result)  # type: ignore
            except ValidationError as err:
                logger.debug(err)
                continue
            updates_list.append(update)
        return updates_list, latest_update_id  # type: ignore

    async def send_message(
        self,
        chat_id: int,
        text: str,
        reply_to_message_id: Optional[int] = None,
        parse_mode: str = "html",
        disable_web_page_preview: bool = False,
    ) -> models.Message:
        result = await self._call_api(
            "sendmessage",
            chat_id=chat_id,
            text=text,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview,
        )
        return models.Message(**result)  # type: ignore

    async def restrict_chat_member(
        self,
        chat_id: int,
        user_id: int,
        permissions: ChatPermissions,
        until_date: datetime,
    ) -> bool:
        result = await self._call_api(
            "restrictChatMember",
            chat_id=chat_id,
            user_id=user_id,
            permissions=permissions.json(),
            until_date=int(until_date.timestamp()),
        )
        return result  # noqa, type: ignore

    async def get_chat_administrators(
        self,
        chat_id: int,
    ) -> List[Union[ChatMemberAdministrator, ChatMemberOwner]]:
        results = await self._call_api(
            "getChatAdministrators",
            chat_id=chat_id,
        )
        admins = []
        for result in results:
            status = result["status"]
            if status == ChatMemberStatuses.CREATOR:
                admins.append(ChatMemberOwner(**result))
            if status == ChatMemberStatuses.ADMINISTRATOR:
                admins.append(ChatMemberAdministrator(**result))
        return admins

    async def pin_chat_message(
        self,
        chat_id: int,
        message_id: int,
        disable_notification: bool = True,
    ) -> bool:
        result = await self._call_api(
            "pinChatMessage",
            chat_id=chat_id,
            message_id=message_id,
            disable_notification=disable_notification,
        )
        return result  # noqa, type: ignore

    async def unpin_chat_message(
        self,
        chat_id: int,
        message_id: int,
    ) -> bool:
        result = await self._call_api(
            "unpinChatMessage",
            chat_id=chat_id,
            message_id=message_id,
        )
        return result  # noqa, type: ignore

    async def unpin_all_chat_messages(
        self,
        chat_id: int,
    ) -> bool:
        result = await self._call_api(
            "unpinAllChatMessages",
            chat_id=chat_id,
        )
        return result  # noqa, type: ignore



from enum import Enum, auto, unique
from typing import Any, Callable, Coroutine, Hashable, NamedTuple, TypedDict, Union

# pylint: disable=invalid-name
HANDLER_TYPING = Union[
    Callable[[Any], None],
    Callable[[Any], Coroutine[Any, Any, Any]],
]


class RegexMatchKwargs(TypedDict):
    matched: tuple[str, ...]


class MatchedToken(NamedTuple):
    token: Hashable
    args: tuple[str, ...] = ()
    kwargs: dict[str, Any] = {}


@unique
class EventTypes(Enum):
    STARTUP = auto()
    SHUTDOWN = auto()
    MESSAGE = auto()



from abc import ABC
from asyncio import iscoroutinefunction
from collections import defaultdict
from inspect import signature
from typing import Any, Callable, Hashable, Iterable, NamedTuple, Optional, Union

from chatushka.core.models import HANDLER_TYPING, MatchedToken
from chatushka.core.protocols import MatcherProtocol
from chatushka.core.transports.models import Update
from chatushka.core.transports.telegram_bot_api import TelegramBotApi


class HelpMessage(NamedTuple):
    tokens: tuple[str]
    message: Optional[str]


class MatcherBase(ABC):
    def __init__(
        self,
    ) -> None:
        self.handlers: dict[Hashable, list[HANDLER_TYPING]] = defaultdict(list)
        self.matchers: list[MatcherProtocol] = []
        self._help_messages: list[HelpMessage] = []

    def __call__(
        self,
        *tokens: Hashable,
        help_message: Optional[str] = None,
        include_in_help: bool = True,
    ) -> Callable[[Callable[[], None]], None]:
        def decorator(
            func: HANDLER_TYPING,
        ) -> None:
            self.add_handler(
                tokens=tokens,
                handler=func,
                help_message=help_message,
                include_in_help=include_in_help,
            )

        return decorator

    @property
    def help_messages(self):
        messages = self._help_messages.copy()
        for matcher in self.matchers:
            messages += matcher.help_messages
        return messages

    def add_handler(
        self,
        tokens: Union[Hashable, Iterable[Hashable]],
        handler: HANDLER_TYPING,
        help_message: Optional[str] = None,
        include_in_help: bool = True,
    ) -> None:
        if not help_message:
            help_message = f"help message of {self.__class__.__name__}"
        if not isinstance(tokens, (list, tuple, set)):
            tokens = (tokens,)
        for raw_token in tokens:
            if isinstance(raw_token, str):
                raw_token = raw_token.strip()
            prepared = self._cast_token(raw_token)
            if not isinstance(prepared, (list, tuple, set)):
                prepared = (prepared,)
            for token in prepared:
                self.handlers[token].append(handler)
        if include_in_help:
            self._help_messages.append(
                HelpMessage(tokens, help_message),
            )

    def add_matcher(
        self,
        *matchers: MatcherProtocol,
    ) -> None:
        self.matchers += matchers

    async def match(
        self,
        api: TelegramBotApi,
        update: Update,
        *,
        should_call_matched: bool = False,
    ) -> list[MatchedToken]:
        matched_handlers = []
        for token in self.handlers.keys():
            if matched := await self._check(token, update):
                matched_handlers.append(matched)
                if should_call_matched:
                    await self.call(
                        api=api,
                        token=matched.token,
                        update=update,
                        kwargs=matched.kwargs | dict(args=matched.args),
                    )
        for matcher in self.matchers:
            matched_handlers += await matcher.match(api, update, should_call_matched=should_call_matched)
        return matched_handlers

    async def call(
        self,
        api: TelegramBotApi,
        token: Hashable,
        update: Optional[Update] = None,
        kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        if not kwargs:
            kwargs = {}
        kwargs = kwargs | dict(api=api, update=update, token=token)
        handlers = self.handlers.get(token)
        if not handlers:
            return
        for handler in handlers:
            sig = signature(handler)
            sig_kwargs = {param: kwargs.get(param) for param in sig.parameters if param in kwargs}
            if update and update.message is not None and "message" in sig.parameters:
                sig_kwargs["message"] = update.message
            if iscoroutinefunction(handler):
                await handler(**sig_kwargs)  # type: ignore
                continue
            handler(**sig_kwargs)

    # pylint: disable=no-self-use
    def _cast_token(
        self,
        token: Hashable,
    ) -> Union[Any, Iterable[Any]]:
        return (token,)  # noqa

    # pylint: disable=unused-argument
    async def _check(
        self,
        token: Hashable,
        update: Update,
    ) -> Optional[MatchedToken]:
        return None

    async def init(self) -> None:
        pass



from chatushka.core.matchers.chat_users_movements import ChatUsersMovementsEventsEnum, ChatUsersMovementsMatcher
from chatushka.core.matchers.commands import CommandsMatcher
from chatushka.core.matchers.cron import CronMatcher
from chatushka.core.matchers.events import EventsMatcher
from chatushka.core.matchers.regex import RegexMatcher
from chatushka.core.models import EventTypes
from chatushka.core.protocols import MatcherProtocol

__all__ = (
    "EventTypes",
    "CommandsMatcher",
    "CronMatcher",
    "EventsMatcher",
    "RegexMatcher",
    "MatcherProtocol",
    "ChatUsersMovementsMatcher",
    "ChatUsersMovementsEventsEnum",
)
